main (argc,argv)
int argc;
char *argv[];
{
/*
	client process to connect to  the time service via port 37 using
	udp.  The client process uses the time message received to check 
	(and optionally to set) the time of the local clock.  the comparison 
	assumes that the local clock keeps time in seconds from 1/1/70 
	which is the UNIX standard, and it assumes that the received time
	is in seconds since 1900.0 to conform to rfc868.  It computes the
	time difference by subtracting 2208988800 from the received time
	to convert it to seconds since 1970 and then makes the comparison
	directly. If the local machine keeps time in some other way, then 
	the comparison method will have to change, but the rest should be 
	okay.

	This software was developed with US Government support
	and it may not be sold, restricted or licensed.  You 
	may duplicate this program provided that this notice
	remains in all of the copies, and you may give it to
	others provided they understand and agree to this
	condition.

	This program and the time protocol it uses are under 
	development and the implementation may change without 
	notice.

	For questions or additional information, contact:

	Judah Levine
	Time and Frequency Division
	NIST/847
	325 Broadway
	Boulder, Colorado 80305
	(303) 492 7785
	jlevine@boulder.nist.gov
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <string.h>
#include <ctype.h>
long int address;		/* holds ip address */
int pserv = 37;		/* time port number on server */
char *cp;			/* server address in . notation */
char *sp;			/* temporary for server address*/
int aindots[4];                 /* numerical host address in dot notation*/
char addrbuf[20];               /* address formatted into dot notation*/
int j;
struct sockaddr_in sin;		/* socket address structure */
int s;				/* socket number */
int length;			/* size of message */
char buf[10];			/* holds message */
unsigned long recdtime;		/* received time in local byte order */
unsigned long dorg = 2208988800ul;	/*seconds from 1900.0 -> 1970.0*/
unsigned long netcons;		/*received time in network order */
struct timeval tvv,tgg;		/* holds local time */
long int diff;			/* time difference, local - NIST */
char cc;
int stat;
int use_serv=0;			/*use server number 1 by default,see below*/
int sw();			/*parses command line switches*/
char let;			/*switch letter*/
long int val;			/*value associated with switch*/
struct hostent *serv0;          /*pointer to structure returned by gethost */
/*
        the following is a list of the servers operated by
        NIST. All of them will support the daytime protocol
        in the format that this program expects.

        Each server may be specified either by name, as in
        time.nist.gov or by ip as in 192.43.244.18. If the
        first character of the specification is a digit, then
        the numerical format is assumed. If a name is entered,
        it is converted to the corresponding ip address using
        the standard DNS query. The program will fail if the
        DNS query cannot find the ip number of the server


        For more information about these servers, look at the
        entry for the Internet Time Service on the Time and
        Frequency Division homepage at www.boulder.nist.gov/timefreq
*/
#define NUMSRV 15
char *serv_ip[NUMSRV]= {"64.236.96.53"   ,/*nist1.aol-va.truetime.com*/
                        "128.138.140.44" ,/*utcnist.colorado.edu*/
                        "207.200.81.113" ,/*nist1.aol-ca.truetime.com*/
                        "216.200.93.8"   ,/*nist1-dc.glassey.com*/
                        "63.149.208.50"  ,/*nist1.datum.com*/
                        "208.184.49.9"   ,/*nist1-ny.glassey.com*/
                        "207.126.103.204",/*nist1-sj.glassey.com*/
                        "129.6.15.28"    ,/*time-a.nist.gov*/
                        "129.6.15.29"    ,/*time-b.nist.gov*/
                        "132.163.4.101"  ,/*time-a.timefreq.bldrdoc.gov*/
                        "132.163.4.102"  ,/*time-b.timefreq.bldrdoc.gov*/
                        "132.163.4.103"  ,/*time-c.timefreq.bldrdoc.gov*/
                        "192.43.244.18"  ,/*time.nist.gov*/
                        "131.107.1.10"   ,/*time-nw.nist.gov*/
			"utcnist.colorado.edu" /*DNS entry = 128.138.140.44*/
			};
/*
	parse command line switch to select server

       -u<j>           use server number j, where j is the index number
                        of the NIST server chosen from the following list.
                        the first server on the list is number 1, the
                        second is number 2, etc.
                        the default is to use the first server, which
                        is number 1.
                        note that the array uses C indexing, so that the
                        first entry in the array is number 0, etc.
                        thus any user response is decremented before
                        being used.

        -u0             the name of the server is given as the next
                        parameter on the command line. the name can
                        be either a name or an ip address in dot
                        notation. The entry will be taken as a name
                        if the first non-blank character is not a
                        digit.
*/
       while( sw(argc,argv,&let,&val) != 0) /*switch is present*/
           {
           switch(let)
              {
              case 'u':
               if(val == 0)    /*next parameter specifies server*/
                   {
                   argc--;
                   argv++;      /*skip over the switch*/
                   if(argc <= 1)
                        {
                        printf("\n Expected server name is missing.\n\n");
                        exit();
                   }
                   cp=argv[1];  /*save the next parameter as the server name*/
                   use_serv=999;        /*set flag*/
                   break;
                }
                /*  server 1 has internal index 0 */
                use_serv=val-1;
                /*  check if entry is out of range*/
                if(use_serv < 0) use_serv=0;
                if(use_serv >= NUMSRV) use_serv=NUMSRV - 1;
                break;
             default:
                fprintf(stderr,"\nSwitch %c not recognized.",let);
                break;
           }
           argc--;      /*decrement argument counter */
           argv++;      /*and increment pointer */
        }
/*
	internet address of selected server
*/
	if(use_serv != 999) cp=serv_ip[use_serv];
        if(!isdigit(*cp))       /*first char not a digit, must convert name*/
           {
           if( (serv0=gethostbyname(cp)) == NULL)
                {
                printf("\n Cannot resolve name %s",cp);
                exit();
           }
           if(serv0->h_length != 4)
              {
              printf("\nLength of host address (= %d) is wrong, 4 expected.",
                        serv0->h_length);
              exit();
           }
           sp= serv0->h_addr_list[0];
           for(j=0; j<4; j++)   /*store and convert address*/
              {
              aindots[j]= *(sp++);
              aindots[j] &= 0xff;       /*turn off sign extension*/
           }
           sprintf(addrbuf,"%d.%d.%d.%d",
                aindots[0],aindots[1],aindots[2],aindots[3]);
	   cp= addrbuf;
        }
        printf("\n Using server at address %s",cp);
/*
	convert address to internal format
*/
	if( (address=inet_addr(cp) ) == -1)
	   {
	   fprintf(stderr,"\n Internet address error.");
	   exit(1);
	}
	bzero( (char *)&sin,sizeof(sin));
	sin.sin_addr.s_addr=address;
	sin.sin_family=AF_INET;
	sin.sin_port=htons(pserv);
/*
	create the socket and then connect to the server
	note that this is a UDP datagram socket 
*/
	if( (s=socket(AF_INET,SOCK_DGRAM,0) ) < 0)
	   {
	   fprintf(stderr,"\n Socket creation error.");
	   exit(1);
	}
	if(connect(s, (struct sockaddr *) &sin, sizeof(sin) ) < 0)
	   {
	   perror(" time server Connect failed --");
	   exit(1);
	}
/*
	send a UDP datagram to start the server 
	going.
*/
	write(s,buf,10);
	length=read(s,&netcons,4);
	gettimeofday(&tvv,0);	/*get time as soon as read completes*/
	if(length <= 0) 
	   {
	   fprintf(stderr,"\n No response from server.");
	   close(s);
	   exit(1);
	}
	recdtime=ntohl(netcons);	/*convert to local byte order */
	recdtime -= dorg;	/*convert to seconds since 1970*/
	close(s);
/*
	now compute difference between local time and
	received time.
*/
	diff=tvv.tv_sec - recdtime;
	if(tvv.tv_usec >= 500000l) diff++;	/*round microseconds */
	printf("\nLocal Clock - NIST = %ld second(s).",diff);
	if(diff == 0)
	   {
	   printf("\n No adjustment is needed.\n");
	   exit();
	}
	printf("\n Do you want to adjust the local clock ? [y/n] ");
	cc=getchar();
	if( (cc == 'y') || (cc == 'Y') )
	   {
	   if(diff > 2100l)
	      {
	      printf("\n Adjustment limited to -2100 s.");
	      diff= 2100l;
	   }
	   if(diff < -2100l)
	      {
	      printf("\n Adjustment limited to +2100 s.");
	     diff= -2100l;
	   }
	   tvv.tv_sec= -diff;
	   tvv.tv_usec = 0;
	   stat=adjtime(&tvv,&tgg);
	   if(stat == 0) printf("\n Adjustment was performed.\n");
	   else perror("Adjustment failed. ");
	}
}
